﻿using System;

// DateTime
namespace GregoryAdam.Base.ExtensionMethods
{

	public static partial class ExtensionMethods
	{
		#region ISO
		//-------------------------------------------------------------------------
		/// <summary>
		/// returns the ISO DayOfWeek 1=Mon, 2=Tue, .... 7=Sun
		/// </summary>
		/// <param name="dow"></param>
		/// <returns></returns>
		public static int DayOfWeekISO(this DayOfWeek dow)
		{
			return dow != DayOfWeek.Sunday ? (int) dow : 7;
		}
		//-------------------------------------------------------------------------
		// week one includes the first Thursday, ie at least 4 days of the week
		// or
		// week one is the week that contains Jan 4
		public static int GetISOWeek(this DateTime dt)
		{
			bool monthJanOrFeb = (dt.Month <= 2);

			int a = monthJanOrFeb ? dt.Year - 1 : dt.Year;
			int b = a / 4 - a / 100 + a / 400;
			int c = (a - 1) / 4 - (a - 1) / 100 + (a - 1) / 400;
			int s = b - c;
			int e = monthJanOrFeb ? 0 : s + 1;
			int f = dt.DayOfYear - 1;

			int g = (a + b) % 7;
			int d = (f + g - e) % 7; // 0 = mon, .... 6 = sun
			int n = f + 3 - d;

			int year, week;

			if (n < 0)
			{
				year = dt.Year - 1;
				week = 53 - (g - s) / 5;
			}
			else if (n > 364 + s)
			{
				year = dt.Year + 1;
				week = 1;
			}
			else
			{
				year = dt.Year;
				week = n / 7 + 1;
			}

			return year * 100 + week;
		}
		//-------------------------------------------------------------------------
		public static int GetNumberOfISOWeeks(this DateTime dt)
		{
			int yearNow, yearPrev;
			yearNow = dt.Year;
			yearPrev = yearNow - 1;
			
			return	(yearNow / 4 - yearNow/100 + yearNow/400) % 7 == 4 
				||	(yearPrev / 4 - yearPrev/100 + yearPrev/400) % 7 == 3 ?
				53 : 52 ;

		}
		public static DateTime ToWeek(this DateTime dt, int week)
		{

			if (week <= 0 || week > dt.GetNumberOfISOWeeks())
			{
				throw new ArgumentOutOfRangeException("week", "outside range");
			}
			// the first week of the year is Jan 04
			// go to Jan 04
			// add (week-1) * 7 days
			// go to the same day of the week
			return dt.AddDays(-dt.DayOfYear + 4 + (week - 1) * 7).ToDayOfWeek(dt.DayOfWeek) ;

		}
		#endregion ISO
		#region Year
		#region ToFirstDayOfYear
		/// <summary>
		/// returns the First DayOfWeek dow of the year
		/// </summary>
		/// <param name="dt"></param>
		/// <param name="dow"></param>
		/// <returns></returns>
		public static DateTime ToFirstDayOfYear(this DateTime dt, DayOfWeek dow)
		{
			// go to end of previous year
			// go to Next DayOfWeek
			return dt.AddDays(-dt.DayOfYear).ToNextDay(dow);
		}
		#endregion
		#region ToLastDayOfYear
		public static DateTime ToLastDayOfYear(this DateTime dt, DayOfWeek dow)
		{
			// go to the end of last year
			// add one year
			// go to lastest DayOfWeek
			return dt.AddDays(-dt.DayOfYear).AddYears(1).ToLatestDay(dow);
		}
		#endregion
		#region GetEaster
		public static DateTime GetEaster(this DateTime dt)
		{
			// http://www.tondering.dk/claus/cal/calendar29.pdf
			int G = dt.Year % 19;
			int C = dt.Year / 100;
			int H = (C - C/4 - (C*8 + 13)/25 + G * 19 + 15) % 30;
			int I = H - H / 28 * (1 - (29/(H+1)) *  (21-G)/11 ) ;
			int J = (dt.Year + dt.Year / 4 + I + 2 - C + C/4) % 7;
			int L = I - J;
			int easterMonth = 3 + (L + 40) / 44;
			int easterDay = L + 28 - 31 * (easterMonth / 4);

			return new DateTime(dt.Year, easterMonth, easterDay);
		}
		#endregion
		#endregion
		#region Month
		#region ToBeginningOfMonth
		//-------------------------------------------------------------------------
		/// <summary>
		/// returns the beginning of the month
		/// </summary>
		/// <param name="dt"></param>
		/// <returns></returns>

		public static DateTime ToBeginningOfMonth(this DateTime dt)
		{
			return dt.AddDays(1.0 - dt.Day);

		}
		//-------------------------------------------------------------------------
		#endregion
		#region ToEndOfMonth
		//-------------------------------------------------------------------------
		public static DateTime ToEndOfMonth(this DateTime dt)
		{
			return dt.ToBeginningOfMonth().AddMonths(1).AddDays(-1.0);
		}
		#endregion
		#region ToFirstDayOfMonth
		/// <summary>
		/// Goes to the first DayOfWeek in the same month
		/// </summary>
		/// <param name="dt"></param>
		/// <param name="dow"></param>
		/// <returns></returns>
		public static DateTime ToFirstDayOfMonth(this DateTime dt, DayOfWeek dow)
		{
			return dt.ToBeginningOfMonth().ToFirstDay(dow);
		}
		#endregion
		#region ToFirstLastOfMonth
		/// <summary>
		/// Goes to the last DayOfWeek in the same month
		/// </summary>
		/// <param name="dt"></param>
		/// <param name="dow"></param>
		/// <returns></returns>
		public static DateTime ToLastDayOfMonth(this DateTime dt, DayOfWeek dow)
		{
			return dt.ToEndOfMonth().ToLatestDay(dow);
		}
		#endregion
		#endregion

		#region Week
		/// <summary>
		/// Goes to a specific Day of week in the same week
		/// Weeks start on Monday and end on Sunday
		/// </summary>
		/// <param name="dt"></param>
		/// <param name="dow"></param>
		/// <returns></returns>
		public static DateTime ToDayOfWeek(this DateTime dt, DayOfWeek dow)
		{
			// go back to previous sunday
			// advance to dayOfWeek requested
			return dt.AddDays(-DayOfWeekISO(dt.DayOfWeek) + DayOfWeekISO(dow));

		}
		/// <summary>
		/// returns the Monday of the week
		/// </summary>
		/// <param name="dt"></param>
		/// <returns></returns>
		public static DateTime ToBeginningOfWeek(this DateTime dt)
		{
			return dt.ToDayOfWeek(DayOfWeek.Monday);
		}
		/// <summary>
		/// returns the Sunday of the week
		/// </summary>
		/// <param name="dt"></param>
		/// <returns></returns>
		public static DateTime ToEndOfWeek(this DateTime dt)
		{
			return dt.ToDayOfWeek(DayOfWeek.Sunday);
		}

		#endregion

		#region Advance or go back to a DayOfWeek
		private static int OffsetMatrixValue(int start, DateTime dt, DayOfWeek offset)
		{
			//(7 - (int)(dt.DayOfWeek) + (int)dow) % 7
			return (start - (int)dt.DayOfWeek + (int)offset) % 7;
		}
		#region ToFirstDay
		/// <summary>
		/// Advances to first DayOfWeek specified.  If already on the DayOfWeek, nothing is done
		/// </summary>
		/// <param name="dt"></param>
		/// <param name="dow"></param>
		/// <returns></returns>
		/// 
		/*
		 *     to   0   1   2   3   4   5   6
		 *      -+----+---+---+---+---+---+---
		 * from 0|  0 | 1 | 2 | 3 | 4 | 5 | 6
		 *      -+----+---+---+---+---+---+---
		 *		1|  6 | 0 | 1 | 2 | 3 | 4 | 5
		 *      -+----+---+---+---+---+---+---
		 *		2|  5 | 6 | 0 | 1 | 2 | 3 | 4
		 *      -+----+---+---+---+---+---+---
		 *		3|  4 | 5 | 6 | 0 | 1 | 2 | 3
		 *      -+----+---+---+---+---+---+---
		 *		4|  3 | 4 | 5 | 6 | 0 | 1 | 2
		 *      -+----+---+---+---+---+---+---
		 *		5|  2 | 3 | 4 | 5 | 6 | 0 | 1
		 *      -+----+---+---+---+---+---+---
		 *		6|  1 | 2 | 3 | 4 | 5 | 6 | 0
		 *      -+----+---+---+---+---+---+---
		 */

		public static DateTime ToFirstDay(this DateTime dt, DayOfWeek dow)
		{
			return dt.AddDays(OffsetMatrixValue(7, dt, dow));
		}
		#endregion
		//-------------------------------------------------------------------------
		#region ToNextDay
		/// <summary>
		/// Advances to next DayOfWeek. If already on the DayOfWeek, a full week is added
		/// </summary>
		/// <param name="dt"></param>
		/// <param name="dow"></param>
		/// <returns></returns>
		/// 
		/*
		 *     to   0   1   2   3   4   5   6
		 *      -+----+---+---+---+---+---+---
		 * from 0|  7 | 1 | 2 | 3 | 4 | 5 | 6
		 *      -+----+---+---+---+---+---+---
		 *		1|  6 | 7 | 1 | 2 | 3 | 4 | 5
		 *      -+----+---+---+---+---+---+---
		 *		2|  5 | 6 | 7 | 1 | 2 | 3 | 4
		 *      -+----+---+---+---+---+---+---
		 *		3|  4 | 5 | 6 | 7 | 1 | 2 | 3
		 *      -+----+---+---+---+---+---+---
		 *		4|  3 | 4 | 5 | 6 | 7 | 1 | 2
		 *      -+----+---+---+---+---+---+---
		 *		5|  2 | 3 | 4 | 5 | 6 | 7 | 1
		 *      -+----+---+---+---+---+---+---
		 *		6|  1 | 2 | 3 | 4 | 5 | 6 | 7
		 *      -+----+---+---+---+---+---+---
		 */
		public static DateTime ToNextDay(this DateTime dt, DayOfWeek dow)
		{
			// return dt.AddDays(1.0).ToFirstDay(dow);
			//return dt.AddDays( ((6 - (int)(dt.DayOfWeek) + (int)dow) % 7) + 1.0);
			return dt.AddDays(OffsetMatrixValue(6, dt, dow) + 1);
		}
		#endregion

		#region ToLastestDay
		/// <summary>
		/// Goes back to latest DayOfWeek specified.  If already on the DayOfWeek, nothing is done
		/// </summary>
		/// <param name="dt"></param>
		/// <param name="dow"></param>
		/// <returns></returns>
		/*
		 * values of matrix(ToNextDay) minus 7 days
		 */
		public static DateTime ToLatestDay(this DateTime dt, DayOfWeek dow)
		{
			// return dt.AddDays(OffsetMatrixValue(6, dt, dow) + 1 - 7 );
			return dt.AddDays(OffsetMatrixValue(6, dt, dow) - 6);
		}
		#endregion

		#region ToPreviousDay
		/// <summary>
		/// Goes back to previous DayOfWeek specified.  If already on the DayOfWeek, a full week is subtracted
		/// </summary>
		/// <param name="dt"></param>
		/// <param name="dow"></param>
		/// <returns></returns>
		/*
		 * values of matrix(ToFirstDay) minus 7 days
		 */
		public static DateTime ToPreviousDay(this DateTime dt, DayOfWeek dow)
		{
			return dt.AddDays(OffsetMatrixValue(7, dt, dow) - 7);
		}
		#endregion
		#endregion
	}
}
